/**
 * Copyright (c) 2009-2012 EBM WebSourcing, 2012-2013 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.bc.sql.listener;

import java.io.IOException;
import java.io.InputStream;
import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.ExecutorService;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.jbi.messaging.MessagingException;
import javax.sql.DataSource;
import javax.xml.transform.stream.StreamSource;

import org.ow2.petals.bc.sql.Constants;
import org.ow2.petals.bc.sql.DataSources;
import org.ow2.petals.bc.sql.SqlComponent;
import org.ow2.petals.bc.sql.service.SQLParameter;
import org.ow2.petals.bc.sql.service.SQLService;
import org.ow2.petals.commons.exception.ExceptionUtil;
import org.ow2.petals.commons.log.PetalsExecutionContext;
import org.ow2.petals.component.framework.api.Message;
import org.ow2.petals.component.framework.api.configuration.ConfigurationExtensions;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.listener.AbstractJBIListener;
import org.ow2.petals.component.framework.util.SourceUtil;
import org.ow2.petals.component.framework.util.XMLUtil;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * @author Adrien Louis - EBM WebSourcing
 */
public class JBIListener extends AbstractJBIListener {

    private DataSources dataSources;

    private SQLService sqlService;

    private ExecutorService threadPool;

    /**
     * Creates a new instance of {@link JBIListener}
     * 
     */
    public JBIListener() {
    }

    @Override
    public void init() {
        this.dataSources = ((SqlComponent) this.getComponent()).getDataSources();
        this.sqlService = ((SqlComponent) this.getComponent()).getSQLService();
        this.threadPool = ((SqlComponent) this.getComponent()).getThreadPool();
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.listener.AbstractJBIListener#onJBIMessage
     * (org.objectweb.petals.component.framework.util.Exchange)
     */
    @Override
    public final boolean onJBIMessage(final Exchange exchange) {
        Connection connection = null;
        if (exchange.isActiveStatus()) {
            try {
                if (!validateExchange(exchange)) {
                    this.getLogger().log(Level.WARNING, exchange.getError().getMessage());
                    // no SQL request, error is set in the exchange
                    return true;
                }

                final String operation = exchange.getOperationName().toLowerCase();
                PetalsExecutionContext.putFlowStepOperationName(exchange.getOperation());
                String sql = null;
                List<SQLParameter> sqlParameters = null;

                if (Constants.OPERATION_STOREDPROCEDURE.equals(operation)
                        || Constants.OPERATION_XML_STOREDPROCEDURE.equals(operation)) {
                    sqlParameters = getStoredProcedureParams(exchange, this.getExtensions(),
                            this.getLogger());
                } else {
                    sql = getSQLRequest(exchange);
                }

                if (exchange.getError() != null) {
                    this.getLogger().log(Level.WARNING, exchange.getError().getMessage());
                    // no SQL or Stored Procedure request, error is set in the
                    // exchange
                    return true;
                }

                // get a JDBC connection
                final DataSource dataSource = this.dataSources.getDataSource(this.getProvides());
                if (dataSource == null) {
                    throw new MessagingException("No dataSource has been found for the service "
                            + exchange.getService());
                }
                // open the connection
                this.getLogger().finest("Get a JDBC connection from the DataSource");
                connection = dataSource.getConnection();

                if (Constants.OPERATION_STOREDPROCEDURE.equals(operation)) {
                    String storedStatementCall = this.getExtensions().get(
                            Constants.STORED_PROCEDURE_CALL);
                    this.sqlService.storedProcedure(sqlParameters, storedStatementCall, connection);

                    // response message is written in a separate thread, while
                    // the
                    // exchange is sent back in the current one
                    final OutStreamer outStreamer = new OutStreamer(this.sqlService, sqlParameters,
                            connection, this.getLogger(), false);
                    this.threadPool.execute(outStreamer);
                    final InputStream inputStream = outStreamer.getPipedInputStream();
                    exchange.setOutMessageContent(new StreamSource(inputStream));
                    connection = null;
                } else if (Constants.OPERATION_XML_STOREDPROCEDURE.equals(operation)) {
                    String storedStatementCall = this.getExtensions().get(
                            Constants.STORED_PROCEDURE_CALL);
                    this.sqlService.storedProcedure(sqlParameters, storedStatementCall, connection);

                    // response message is written in a separate thread, while
                    // the
                    // exchange is sent back in the current one
                    final OutStreamer outStreamer = new OutStreamer(this.sqlService, sqlParameters,
                            connection, this.getLogger(), true);
                    this.threadPool.execute(outStreamer);
                    final InputStream inputStream = outStreamer.getPipedInputStream();
                    exchange.setOutMessageContent(new StreamSource(inputStream));
                    connection = null;
                } else if (Constants.OPERATION_SELECT.equals(operation)) {
                    final ResultSet resultSet = this.sqlService.select(sql, connection);

                    final boolean showMetadata = this.getShowMetadata(exchange);
                    if (showMetadata) {
                        this.getLogger().finest("Metadata written in the response");
                    }
                    // response message is written in a separate thread, while
                    // the
                    // exchange is sent back in the current one
                    final OutStreamer outStreamer = new OutStreamer(this.sqlService, resultSet,
                            showMetadata, connection, this.getLogger());
                    this.threadPool.execute(outStreamer);
                    final InputStream inputStream = outStreamer.getPipedInputStream();
                    exchange.setOutMessageContent(new StreamSource(inputStream));
                    connection = null;
                } else if (Constants.OPERATION_INSERT.equals(operation)
                        || Constants.OPERATION_UPDATE.equals(operation)
                        || Constants.OPERATION_DELETE.equals(operation)) {
                    this.sqlService.update(sql, connection);
                } else {
                    final int result = this.sqlService.update(sql, connection);
                    final String resultXML = "<updated xmlns=\"http://petals.ow2.org/components/sql/version-1\">"
                            + result + "</updated>";
                    exchange.getOutMessage().setContent(SourceUtil.createSource(resultXML));
                }

            } catch (MessagingException e) {
                this.getLogger().log(Level.WARNING, e.getMessage(), e);
                exchange.setError(e);
            } catch (PEtALSCDKException e) {
                this.getLogger().log(Level.WARNING, e.getMessage(), e);
                exchange.setError(new MessagingException(ExceptionUtil.getExtendedMessage(e)));
            } catch (IOException e) {
                this.getLogger().log(Level.WARNING, e.getMessage(), e);
                exchange.setError(e);
            } catch (SQLException e) {
                this.getLogger().log(Level.WARNING, e.getMessage(), e);
                if (e.getClass().getName().startsWith("java.sql")) {
                    exchange.setError(e);
                } else {
                    // the exception is defined in a JDBC library, can not be
                    // propagate directly
                    exchange.setError(new MessagingException(ExceptionUtil.getExtendedMessage(e)));
                }
            } catch (ParseException e) {
                this.getLogger().log(Level.WARNING, e.getMessage(), e);
                exchange.setError(e);
            } finally {
                // close the connection (except for streaming mode)
                if (connection != null) {
                    try {
                        connection.close();
                    } catch (final SQLException e) {
                        this.getLogger().log(Level.WARNING, e.getMessage(), e);
                    }
                }
            }
        }

        return true;
    }

    /**
     * Get the SQL request from the exchange. If failed, an error is set to the
     * exchange.
     * 
     * @param document
     * @return
     * @throws MessagingException
     */
    private static final String getSQLRequest(final Exchange exchange) throws MessagingException {
        String sql = null;
        final Element queryElement = exchange.getInMessageContentAsDocument().getDocumentElement();
        if (!(Constants.SQL_ELEMENT.equals(queryElement.getLocalName()) && Constants.NAMESPACE
                .equals(queryElement.getNamespaceURI()))) {
            exchange.setError(new MessagingException(
                    "The request message must contain a root element named '"
                            + Constants.SQL_ELEMENT + "' in the namespace '" + Constants.NAMESPACE
                            + "'"));
        } else {
            sql = XMLUtil.getTextContent(queryElement);
        }

        return sql;
    }

    /**
     * Get the map of the IN, INOUT, OUT parameters for a SQL stored procedure.
     * If failed, an error is set to the exchange.
     * 
     * @param document
     * @return
     * @throws MessagingException
     */
    private static final List<SQLParameter> getStoredProcedureParams(final Exchange exchange,
            ConfigurationExtensions extensions, Logger logger) throws MessagingException {
        List<SQLParameter> sqlParameters = new ArrayList<SQLParameter>();
        Short paramRank;
        String externalizedSQLParameter;

        final Element storedStatementElement = exchange.getInMessageContentAsDocument()
                .getDocumentElement();
        if (!(Constants.STORED_PROCEDURE.equals(storedStatementElement.getLocalName()) && Constants.NAMESPACE
                .equals(storedStatementElement.getNamespaceURI()))) {
            exchange.setError(new MessagingException(
                    "The request message must contain a root element named '"
                            + Constants.STOREDPROCEDURE_ELEMENT + "' in the namespace '"
                            + Constants.NAMESPACE + "'"));
        } else {
            final NodeList nodeList = storedStatementElement.getChildNodes();

            for (int i = 0; i < nodeList.getLength(); i++) {
                paramRank = null;

                Node parameterNode = nodeList.item(i);
                if (Node.ELEMENT_NODE == parameterNode.getNodeType()) {

                    if (Constants.PARAMETER_ELEMENT.equals(parameterNode.getLocalName())) {
                        NamedNodeMap attributes = parameterNode.getAttributes();
                        Node rankAttribute = attributes.getNamedItem(Constants.RANK_ATTRIBUTE);
                        if (rankAttribute != null) {
                            try {
                                paramRank = Short.valueOf(rankAttribute.getTextContent());
                            } catch (NumberFormatException e) {
                                // do nothing
                            }
                            if (paramRank != null) {
                                externalizedSQLParameter = extensions.get(Constants.PARAMETER
                                        + paramRank);
                                if (externalizedSQLParameter == null) {
                                    exchange.setError(new MessagingException(
                                            "The request message contains a parameter with an unbound rank : "
                                                    + paramRank));
                                    return null;
                                }
                                SQLParameter sqlParameter = new SQLParameter(
                                        externalizedSQLParameter, parameterNode.getTextContent());
                                if (SQLParameter.OUT_SCOPE.equals(sqlParameter.getScope())) {
                                    exchange.setError(new MessagingException(
                                            "The request message contains an illegal OUT parameter at rank : "
                                                    + paramRank));
                                    return null;
                                }
                                sqlParameters.add(sqlParameter);
                                logger.finest("Found a SQL stored procedure parameter from the request message : "
                                        + sqlParameter);
                            }
                        }
                    }

                    if (paramRank == null) {
                        exchange.setError(new MessagingException(
                                "The request message must contain only parameter elements named '"
                                        + Constants.PARAMETER_ELEMENT
                                        + "' with integer attribute named '"
                                        + Constants.RANK_ATTRIBUTE + "' in the namespace '"
                                        + Constants.NAMESPACE + "'"));
                        return null;
                    }
                }
            }
        }

        paramRank = 1;
        boolean found;
        while ((externalizedSQLParameter = extensions.get(Constants.PARAMETER + paramRank)) != null) {
            found = false;

            for (SQLParameter sqlParameter : sqlParameters) {
                if (sqlParameter.getRank() == paramRank) {
                    found = true;
                    break;
                }
            }
            if (found) {
                paramRank++;
                continue;
            }

            SQLParameter outSqlParameter = new SQLParameter(externalizedSQLParameter, "");
            if (!"OUT".equals(outSqlParameter.getScope())) {
                exchange.setError(new MessagingException("The parameter with rank " + paramRank
                        + " has a " + outSqlParameter.getScope()
                        + " scope and must be set in the request message"));
                return null;
            }
            paramRank++;
            sqlParameters.add(outSqlParameter);
        }

        return sqlParameters;
    }

    /**
     * Is Show Metadata activated?
     * 
     * @param exchange
     * @return
     */
    private final boolean getShowMetadata(final Exchange exchange) {
        boolean showMetadata = false;
        if (exchange.getProperty(Constants.METADATA) != null
                && "true".equals(exchange.getProperty(Constants.METADATA))) {
            showMetadata = true;
        } else if (this.getExtensions().get(Constants.METADATA, "false").equals("true")) {
            showMetadata = true;
        }
        try {
            if (exchange.getInMessageProperty(Constants.METADATA) != null
                    && "true".equals(exchange.getInMessageProperty(Constants.METADATA))) {
                showMetadata = true;
            }
        } catch (final MessagingException e) {
            // do nothing
        }

        return showMetadata;
    }

    /**
     * Validate the operation and the MEP.
     * 
     * @param operation
     */
    private static final boolean validateExchange(final Exchange exchange) {
        boolean result = true;

        final String operation = exchange.getOperationName().toLowerCase();
        if (Constants.OPERATION_SELECT.equals(operation)) {
            if (!Message.MEPConstants.IN_OUT_PATTERN.equals(exchange.getPattern())) {
                exchange.setError(new MessagingException("Operation '" + operation
                        + "' must use the pattern '" + Message.MEPConstants.IN_OUT_PATTERN.value()
                        + "'"));
                result = false;
            }
        } else if (Constants.OPERATION_INSERT.equals(operation)) {
            if (!Message.MEPConstants.IN_ONLY_PATTERN.equals(exchange.getPattern())) {
                exchange.setError(new MessagingException("Operation '" + operation
                        + "' must use the pattern '" + Message.MEPConstants.IN_ONLY_PATTERN.value()
                        + "'"));
                result = false;
            }
        } else if (Constants.OPERATION_INSERT_WITH_RESPONSE.equals(operation)) {
            if (!Message.MEPConstants.IN_OUT_PATTERN.equals(exchange.getPattern())) {
                exchange.setError(new MessagingException("Operation '" + operation
                        + "' must use the pattern '" + Message.MEPConstants.IN_OUT_PATTERN.value()
                        + "'"));
                result = false;
            }
        } else if (Constants.OPERATION_UPDATE.equals(operation)) {
            if (!Message.MEPConstants.IN_ONLY_PATTERN.equals(exchange.getPattern())) {
                exchange.setError(new MessagingException("Operation '" + operation
                        + "' must use the pattern '" + Message.MEPConstants.IN_ONLY_PATTERN.value()
                        + "'"));
                result = false;
            }
        } else if (Constants.OPERATION_UPDATE_WITH_RESPONSE.equals(operation)) {
            if (!Message.MEPConstants.IN_OUT_PATTERN.equals(exchange.getPattern())) {
                exchange.setError(new MessagingException("Operation '" + operation
                        + "' must use the pattern '" + Message.MEPConstants.IN_OUT_PATTERN.value()
                        + "'"));
                result = false;
            }
        } else if (Constants.OPERATION_DELETE.equals(operation)) {
            if (!Message.MEPConstants.IN_ONLY_PATTERN.equals(exchange.getPattern())) {
                exchange.setError(new MessagingException("Operation '" + operation
                        + "' must use the pattern '" + Message.MEPConstants.IN_ONLY_PATTERN.value()
                        + "'"));
                result = false;
            }
        } else if (Constants.OPERATION_DELETE_WITH_RESPONSE.equals(operation)) {
            if (!Message.MEPConstants.IN_OUT_PATTERN.equals(exchange.getPattern())) {
                exchange.setError(new MessagingException("Operation '" + operation
                        + "' must use the pattern '" + Message.MEPConstants.IN_OUT_PATTERN.value()
                        + "'"));
                result = false;
            }
        } else if (Constants.OPERATION_STOREDPROCEDURE.equals(operation)) {
            if (!Message.MEPConstants.IN_OUT_PATTERN.equals(exchange.getPattern())) {
                exchange.setError(new MessagingException("Operation '" + operation
                        + "' must use the pattern '" + Message.MEPConstants.IN_OUT_PATTERN.value()
                        + "'"));
                result = false;
            }
        } else if (Constants.OPERATION_XML_STOREDPROCEDURE.equals(operation)) {
            if (!Message.MEPConstants.IN_OUT_PATTERN.equals(exchange.getPattern())) {
                exchange.setError(new MessagingException("Operation '" + operation
                        + "' must use the pattern '" + Message.MEPConstants.IN_OUT_PATTERN.value()
                        + "'"));
                result = false;
            }
        } else {
            exchange.setError(new MessagingException("Operation '" + operation
                    + "' must have a valid local name.  Choices are :'"
                    + Constants.OPERATION_SELECT + "','" + Constants.OPERATION_INSERT + "','"
                    + Constants.OPERATION_INSERT_WITH_RESPONSE + "','" + Constants.OPERATION_UPDATE
                    + Constants.OPERATION_UPDATE_WITH_RESPONSE + "','" + Constants.OPERATION_DELETE
                    + "','" + Constants.OPERATION_DELETE_WITH_RESPONSE + "','"
                    + Constants.OPERATION_STOREDPROCEDURE + "','"
                    + Constants.OPERATION_XML_STOREDPROCEDURE + "'"));
            result = false;
        }
        return result;
    }
}
