/**
 * Copyright (c) 2009-2012 EBM WebSourcing, 2012-2013 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.bc.sql.listener;

import java.io.IOException;
import java.io.PipedInputStream;
import java.io.PipedOutputStream;
import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.ow2.petals.bc.sql.Constants;
import org.ow2.petals.bc.sql.service.SQLParameter;
import org.ow2.petals.bc.sql.service.SQLService;

/**
 * The SQL Response is written in a PipeStream an thread while the response of
 * is sent by the JBIListener thread.
 * 
 * @author Adrien Louis - EBM WebSourcing
 */
public class OutStreamer implements Runnable {

    private final PipedOutputStream pos = new PipedOutputStream();

    private final PipedInputStream pis = new PipedInputStream();

    private final ResultSet resultSet;

    private final List<SQLParameter> sqlParameters;

    private final SQLService sqlService;

    private final boolean showMetadata;

    private final Logger logger;

    private final Connection connection;
    
    private final boolean isXMLResult;

    /**
     * A Thread used to push resultSet data into a Pipe Stream after the
     * SqlComponent returns the response
     * 
     * @param sqlService
     * @param resultSet
     *            resultSet to transform into an XML doc
     * @param showMetadata
     *            row index, column name and type written as xml attributes
     * @param connection
     *            this connection is closed when all the resultSet has been
     *            written
     * @param logger
     * @throws IOException
     */
    public OutStreamer(SQLService sqlService, ResultSet resultSet, boolean showMetadata,
            Connection connection, Logger logger) throws IOException {
        super();
        this.sqlService = sqlService;
        this.showMetadata = showMetadata;
        this.logger = logger;
        this.resultSet = resultSet;
        this.sqlParameters = null;
        this.connection = connection;
        this.isXMLResult = false;
        this.pos.connect(this.pis);
    }

    /**
     * A Thread used to push resultSet data into a Pipe Stream after the
     * SqlComponent returns the response
     * 
     * @param sqlService
     * @param sqlParameters
     *            SQL parameters to transform into an XML doc
     * @param connection
     *            this connection is closed when all the resultSet has been
     *            written
     * @param logger
     * @throws IOException
     */
    public OutStreamer(SQLService sqlService, List<SQLParameter> sqlParameters,
            Connection connection, Logger logger, boolean isXMLResult) throws IOException {
        super();
        this.sqlService = sqlService;
        this.logger = logger;
        this.sqlParameters = sqlParameters;
        this.resultSet = null;
        this.connection = connection;
        this.pos.connect(this.pis);
        this.isXMLResult = isXMLResult;

        this.showMetadata = false;
    }

    public PipedInputStream getPipedInputStream() {
        return this.pis;
    }

    public void run() {
        try {
            if (this.resultSet != null) {
                this.logger.finest("Build the response of the '" + Constants.OPERATION_SELECT
                        + "' operation");
                this.sqlService.resultSetAsXMLOutputStream(this.resultSet, this.pos,
                        this.showMetadata);
            } else {
                if (this.isXMLResult) {
                this.logger.finest("Build the response of the '"
                        + Constants.OPERATION_XML_STOREDPROCEDURE + "' operation");
                } else {
                    this.logger.finest("Build the response of the '"
                            + Constants.OPERATION_STOREDPROCEDURE + "' operation");
                }
                this.sqlService.sqlParametersAsXMLOutputStream(this.sqlParameters, this.pos, this.isXMLResult);
            }
            this.pos.close();

        } catch (final Exception e) {
            this.logger.log(Level.SEVERE, e.getMessage(), e);
        } finally {
            // close the connection
            try {
                this.connection.close();
            } catch (final SQLException e) {
                this.logger.log(Level.SEVERE, e.getMessage(), e);
            }
        }

    }
}
