/**
 * Copyright (c) 2009-2012 EBM WebSourcing, 2012-2013 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.bc.sql.service;

import java.io.IOException;
import java.io.Reader;
import java.sql.CallableStatement;
import java.sql.Date;
import java.sql.SQLException;
import java.sql.Time;
import java.sql.Timestamp;
import java.sql.Types;
import java.text.DateFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;

import javax.jbi.messaging.MessagingException;

/**
 * @author Roland Naudin - EBM WebSourcing
 */
public class SQLParameter {

    public static final Map<String, Integer> SQL_TYPES = new HashMap<String, Integer>();

    public static final String DEFAULT_SCOPE = "IN";

    public static final String IN_SCOPE = "IN";

    public static final String OUT_SCOPE = "OUT";

    public static final String INOUT_SCOPE = "INOUT";

    private static final DateFormat dateFormat = new SimpleDateFormat("yyyy-MM-dd");

    private static final DateFormat timeFormat = new SimpleDateFormat("HH:mm:ss");

    private static final DateFormat timestampFormat = new SimpleDateFormat(
            "yyyy-MM-dd HH:mm:ss:SSSS");

    static {
        SQL_TYPES.put("ARRAY", Types.INTEGER);
        SQL_TYPES.put("BIGINT", Types.BIGINT);
        SQL_TYPES.put("BINARY", Types.BINARY);
        SQL_TYPES.put("BIT", Types.BIT);
        SQL_TYPES.put("BLOB", Types.BLOB);
        SQL_TYPES.put("BOOLEAN ", Types.BOOLEAN);
        SQL_TYPES.put("CHAR", Types.CHAR);
        SQL_TYPES.put("CLOB", Types.CLOB);
        SQL_TYPES.put("DATALINK", Types.DATALINK);
        SQL_TYPES.put("DATE", Types.DATE);
        SQL_TYPES.put("DECIMAL", Types.DECIMAL);
        SQL_TYPES.put("DISTINCT", Types.DISTINCT);
        SQL_TYPES.put("DOUBLE", Types.DOUBLE);
        SQL_TYPES.put("FLOAT", Types.FLOAT);
        SQL_TYPES.put("INT", Types.INTEGER);
        SQL_TYPES.put("INTEGER", Types.INTEGER);
        SQL_TYPES.put("LONGVARBINARY", Types.LONGVARBINARY);
        SQL_TYPES.put("LONGVARCHAR", Types.LONGVARCHAR);
        SQL_TYPES.put("NUMERIC", Types.NUMERIC);
        SQL_TYPES.put("OTHER", Types.OTHER);
        SQL_TYPES.put("REAL", Types.REAL);
        SQL_TYPES.put("REF", Types.REF);
        SQL_TYPES.put("SMALLINT", Types.SMALLINT);
        SQL_TYPES.put("STRUCT", Types.STRUCT);
        SQL_TYPES.put("TIME", Types.TIME);
        SQL_TYPES.put("TIMESTAMP", Types.TIMESTAMP);
        SQL_TYPES.put("TINYINT", Types.TINYINT);
        SQL_TYPES.put("VARBINARY", Types.VARBINARY);
        SQL_TYPES.put("VARCHAR", Types.VARCHAR);
        SQL_TYPES.put("VARCHAR2", Types.VARCHAR);
        SQL_TYPES.put("LONG", Types.VARCHAR);
    }

    final private int type;

    final private String scope;

    final private short rank;

    private String value;

    public SQLParameter(String type, String scope, short rank) throws MessagingException {
        super();
        this.type = SQL_TYPES.get(type);
        this.scope = scope;
        this.rank = rank;

        if (this.type == 0) {
            throw new MessagingException("SQL Type " + type + "not recognized");
        }
    }

    public SQLParameter(String externalizedSQLParameter, String value) throws MessagingException {

        String[] splittedStrings = externalizedSQLParameter.split(":");
        this.rank = Short.valueOf(splittedStrings[0]);
        this.scope = splittedStrings[1];
        this.type = Integer.valueOf(splittedStrings[2]);
        this.value = value;

        if (this.type == 0) {
            throw new MessagingException("SQL Type " + type + "not recognized");
        }
    }

    public int getType() {
        return type;
    }

    public String getScope() {
        return scope;
    }

    public short getRank() {
        return this.rank;
    }

    public String getValue() {
        return value;
    }

    public void setValue(String value) {
        this.value = value;
    }

    @Override
    public String toString() {
        return this.rank + ":" + this.scope + ":" + this.type;
    }

    /**
     * Set the SQL parameter to the {@link CallableStatement} according to its
     * type. <br>
     * TODO: handle the stream and exotic types
     * 
     * @param statement
     * @param sqlParameter
     * @param rank
     * 
     * @throws SQLException
     * @throws ParseException
     * @throws MessagingException
     * @throws IOException
     */
    public static final void setSQLParameter(CallableStatement statement,
            SQLParameter sqlParameter, short rank) throws SQLException, ParseException,
            MessagingException, IOException {

        switch (sqlParameter.getType()) {
            case (Types.BOOLEAN):
                statement.setBoolean(Short.valueOf(rank), Boolean.valueOf(sqlParameter.getValue()));
                break;
            case (Types.DATE):
                java.util.Date date = dateFormat.parse(sqlParameter.getValue());
                statement.setDate(Short.valueOf(rank), new Date(date.getTime()));
                break;
            case (Types.DOUBLE):
                statement.setDouble(Short.valueOf(rank), Double.valueOf(sqlParameter.getValue()));
                break;
            case (Types.FLOAT):
                statement.setFloat(Short.valueOf(rank), Float.valueOf(sqlParameter.getValue()));
                break;
            case (Types.INTEGER):
                statement.setInt(Short.valueOf(rank), Integer.valueOf(sqlParameter.getValue()));
                break;
            case (Types.LONGVARCHAR):
                statement.setString(Short.valueOf(rank), sqlParameter.getValue());
                break;
            case (Types.SMALLINT):
                statement.setInt(Short.valueOf(rank), Integer.valueOf(sqlParameter.getValue()));
                break;
            case (Types.TIME):
                java.util.Date time = timeFormat.parse(sqlParameter.getValue());
                statement.setTime(Short.valueOf(rank), new Time(time.getTime()));
                break;
            case (Types.TIMESTAMP):
                java.util.Date timestamp = timestampFormat.parse(sqlParameter.getValue());
                statement.setTimestamp(Short.valueOf(rank), new Timestamp(timestamp.getTime()));
                break;
            case (Types.TINYINT):
                statement.setInt(Short.valueOf(rank), Integer.valueOf(sqlParameter.getValue()));
                break;
            case (Types.VARCHAR):
                statement.setString(Short.valueOf(rank), sqlParameter.getValue());
                break;
            case (Types.CLOB):
                Reader clobReader = statement.getClob(rank).getCharacterStream();
                StringBuilder resultBuilder = new StringBuilder();
                int size = 0;
                char[] buffer = new char[64];
                while ((size = clobReader.read(buffer)) != -1) {
                    resultBuilder.append(buffer, 0, size);
                }
                clobReader.close();
                statement.setString(Short.valueOf(rank), resultBuilder.toString());
                break;
            default:
                for (Entry<String, Integer> entry : SQL_TYPES.entrySet()) {
                    if (entry.getValue() == sqlParameter.getType()) {
                        throw new MessagingException("Type : " + entry.getKey()
                                + " is not supported by the SQL component");
                    }
                }

        }
    }

    /**
     * Get the SQL parameter from the {@link CallableStatement} according to its
     * type. <br>
     * TODO: handle the stream and exotic types
     * 
     * @param statement
     * @param sqlParameter
     * @param rank
     * 
     * @throws SQLException
     * @throws IOException
     * @throws MessagingException
     */
    public static final void getSQLParameter(CallableStatement statement,
            SQLParameter sqlParameter, short rank) throws SQLException, IOException,
            MessagingException {
        String result = null;

        switch (sqlParameter.getType()) {
            case (Types.BOOLEAN):
                result = Boolean.toString(statement.getBoolean(rank));
                break;
            case (Types.DATE):
                result = statement.getDate(rank).toString();
                break;
            case (Types.DOUBLE):
                result = Double.toString(statement.getDouble(rank));
                break;
            case (Types.FLOAT):
                result = Float.toString(statement.getFloat(rank));
                break;
            case (Types.INTEGER):
                result = Integer.toString(statement.getInt(rank));
                break;
            case (Types.LONGVARCHAR):
                result = statement.getString(rank);
                break;
            case (Types.SMALLINT):
                result = Integer.toString(statement.getInt(rank));
                break;
            case (Types.TIME):
                result = statement.getTime(rank).toString();
                break;
            case (Types.TIMESTAMP):
                result = statement.getTimestamp(rank).toString();
                break;
            case (Types.TINYINT):
                result = Integer.toString(statement.getInt(rank));
                break;
            case (Types.VARCHAR):
                result = statement.getString(rank);
                break;
            case (Types.CLOB):
                Reader clobReader = statement.getClob(rank).getCharacterStream();
                StringBuilder resultBuilder = new StringBuilder();
                int size = 0;
                char[] buffer = new char[64];
                while ((size = clobReader.read(buffer)) != -1) {
                    resultBuilder.append(buffer, 0, size);
                }
                clobReader.close();
                result = resultBuilder.toString();
                break;
            default:
                for (Entry<String, Integer> entry : SQL_TYPES.entrySet()) {
                    if (entry.getValue() == sqlParameter.getType()) {
                        throw new MessagingException("Type : " + entry.getKey()
                                + " is not supported by the SQL component");
                    }
                }
        }

        sqlParameter.setValue(result);
    }

}
