/**
 * Copyright (c) 2009-2012 EBM WebSourcing, 2012-2013 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.bc.sql.service;

import java.io.IOException;
import java.io.OutputStream;
import java.sql.CallableStatement;
import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;
import java.sql.Statement;
import java.text.ParseException;
import java.util.Iterator;
import java.util.List;
import java.util.logging.Logger;

import javax.jbi.messaging.MessagingException;
import javax.xml.namespace.QName;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamWriter;

import org.ow2.petals.bc.sql.Constants;
import org.ow2.petals.bc.sql.SqlProvideExtFlowStepBeginLogData;
import org.ow2.petals.commons.log.FlowAttributes;
import org.ow2.petals.commons.log.Level;
import org.ow2.petals.commons.log.PetalsExecutionContext;
import org.ow2.petals.component.framework.logger.ProvideExtFlowStepEndLogData;

import com.ebmwebsourcing.easycommons.thread.ExecutionContext;
import com.ebmwebsourcing.easycommons.xml.XMLOutputFactories;

/**
 * @author Adrien Louis - EBM WebSourcing
 */
public class SQLService {

    private final Logger logger;

    public SQLService(Logger logger) {
        super();
        this.logger = logger;
    }

    /**
     * Execute the select SQL command on the given connection.
     * 
     * @param sql
     * @param connection
     * @return
     * @throws SQLException
     */
    public ResultSet select(String sql, Connection connection) throws SQLException {
        Statement statement = connection.createStatement();
        logger.info("Execute the following SQL statement : " + sql);
        addMonitBeginStepExtTrace();
        ResultSet resultSet = statement.executeQuery(sql);
        addMonitEndStepExtTrace();
        return resultSet;
    }

    /**
     * Execute the SQL Stored Procedure on the given connection.
     * 
     * @param sql
     * @param connection
     * @return
     * @throws SQLException
     * @throws ParseException
     * @throws MessagingException
     * @throws IOException
     */
    public void storedProcedure(List<SQLParameter> sqlParameters, String call, Connection connection)
            throws SQLException, ParseException, MessagingException, IOException {
        CallableStatement statement = connection.prepareCall(call);

        for (SQLParameter sqlParameter : sqlParameters) {
            if (SQLParameter.IN_SCOPE.equals(sqlParameter.getScope())
                    || SQLParameter.INOUT_SCOPE.equals(sqlParameter.getScope())) {
                logger.finest("Register a SQL stored procedure parameter : " + sqlParameter);
                SQLParameter.setSQLParameter(statement, sqlParameter, sqlParameter.getRank());
            }

            if (!SQLParameter.IN_SCOPE.equals(sqlParameter.getScope())) {
                logger.finest("Register SQL stored procedure parameter to return : " + sqlParameter);
                statement.registerOutParameter(sqlParameter.getRank(), sqlParameter.getType());
            }
        }

        logger.info("Execute the following SQL stored procedure : " + call);
        addMonitBeginStepExtTrace();
        statement.execute();
        addMonitEndStepExtTrace();

        for (SQLParameter sqlParameter : sqlParameters) {
            if (SQLParameter.OUT_SCOPE.equals(sqlParameter.getScope())
                    || SQLParameter.INOUT_SCOPE.equals(sqlParameter.getScope())) {
                SQLParameter.getSQLParameter(statement, sqlParameter, sqlParameter.getRank());
                logger.finest("Set a returned SQL stored procedure parameter : " + sqlParameter);

            }
        }

    }

    /**
     * Execute the update SQL command on the given connection.
     * 
     * @param sql
     * @param connection
     * @return
     * @throws SQLException
     */
    public int update(String sql, Connection connection) throws SQLException {
        Statement statement = connection.createStatement();
        logger.info("Execute the following SQL instruction : " + sql);
        addMonitBeginStepExtTrace();
        int nbOfUpdatedRows = statement.executeUpdate(sql);
        addMonitEndStepExtTrace();
        return nbOfUpdatedRows;
    }

    /**
     * Transform the given resultSet as an XML outputStream
     * 
     * @param resultSet
     *            the jdbc resultSet to transform
     * @param outputStream
     *            the stream where the XML is written
     * @param showAttributes
     *            row index, column name and type are set as xml attributes in
     *            the response
     * @throws SQLException
     * @throws XMLStreamException
     */
    public void resultSetAsXMLOutputStream(ResultSet resultSet, OutputStream outputStream,
            boolean showAttributes) throws SQLException, XMLStreamException {
        XMLStreamWriter writer = XMLOutputFactories.createXMLStreamWriter(outputStream);

        writer.writeStartDocument();
        writer.writeStartElement(Constants.RESULT_ELEMENT);
        writer.writeDefaultNamespace(Constants.NAMESPACE);

        if (resultSet != null) {
            ResultSetMetaData metaData = resultSet.getMetaData();
            int columnCount = metaData.getColumnCount();
            // resultSet.beforeFirst();
            while (resultSet.next()) {
                writer.writeStartElement(Constants.ROW_ELEMENT);
                if (showAttributes) {
                    writer.writeAttribute(Constants.INDEX_ATTRIBUTE,
                            Integer.toString(resultSet.getRow()));
                }
                for (int i = 1; i <= columnCount; i++) {
                    writer.writeStartElement(Constants.COLUMN_ELEMENT);
                    if (showAttributes) {
                        writer.writeAttribute(Constants.NAME_ATTRIBUTE, metaData.getColumnName(i));
                        writer.writeAttribute(Constants.TYPE_ATTRIBUTE,
                                metaData.getColumnTypeName(i));
                    }
                    final String resultvalue = resultSet.getString(i);
                    if (resultvalue == null) {
                        writer.writeCharacters("NULL");
                    } else {
                        writer.writeCharacters(resultvalue);
                    }
                    writer.writeEndElement();
                }
                writer.writeEndElement();
            }
        }
        writer.writeEndElement();
        writer.writeEndDocument();
        writer.flush();
        writer.close();
    }

    /**
     * Inject the given list of SQL parameters into a XML outputStream.
     * 
     * @param sqlParameters
     *            the list of SQL parameters from the Stored Statement
     * @param outputStream
     *            the stream where the XML is written
     * @param isResultXML
     *            if {@code true},
     * 
     * @throws XMLStreamException
     * @throws IOException
     */
    public void sqlParametersAsXMLOutputStream(List<SQLParameter> sqlParameters,
            OutputStream outputStream, boolean isResultXML) throws XMLStreamException, IOException {
        XMLStreamWriter writer = XMLOutputFactories.createXMLStreamWriter(outputStream);

        writer.writeStartDocument();
        writer.writeStartElement(Constants.STOREDPROCEDURE_RESULT_ELEMENT);
        writer.writeDefaultNamespace(Constants.NAMESPACE);

        // process XMLStoredProcedure response message
        if (isResultXML) {
            for (SQLParameter sqlParameter : sqlParameters) {
                if (SQLParameter.INOUT_SCOPE.equals(sqlParameter.getScope())
                        || SQLParameter.OUT_SCOPE.equals(sqlParameter.getScope())) {
                    writer.writeStartElement(Constants.PARAMETER_ELEMENT);
                    writer.writeAttribute(Constants.RANK_ATTRIBUTE,
                            Short.toString(sqlParameter.getRank()));
                    if (sqlParameter.getValue() == null) {
                        writer.writeCharacters("NULL");
                    } else {
                        writer.writeCharacters(sqlParameter.getValue());
                    }
                    writer.writeEndElement();
                }
            }

        } else {
            // process StoredProcedure response message
            final Iterator<SQLParameter> sqlParameterIterator = sqlParameters.iterator();
            while (sqlParameterIterator.hasNext()) {
                final SQLParameter sqlParameter = sqlParameterIterator.next();
                if (SQLParameter.INOUT_SCOPE.equals(sqlParameter.getScope())
                        || SQLParameter.OUT_SCOPE.equals(sqlParameter.getScope())) {
                    writer.writeCharacters(sqlParameter.getRank() + ":" + sqlParameter.getScope()
                            + ":" + sqlParameter.getValue());
                }
                if (sqlParameterIterator.hasNext()) {
                    writer.writeCharacters(";");
                }
            }
        }

        writer.writeEndElement();
        writer.writeEndDocument();
    }

    private void addMonitBeginStepExtTrace() {
        FlowAttributes flowAttributes = PetalsExecutionContext.getFlowAttributes();
        String flowInstanceId = "";
        String flowStepId = "";
        if (flowAttributes != null) {
            flowInstanceId = flowAttributes.getFlowInstanceId();
            flowStepId = flowAttributes.getFlowStepId();
        }
        String sqlOperation = (String) ExecutionContext.getProperties().get(
                PetalsExecutionContext.FLOW_STEP_OPERATION_NAME_PROPERTY_NAME);
        logger.log(Level.MONIT, "", new SqlProvideExtFlowStepBeginLogData(flowInstanceId,
                flowStepId, sqlOperation));
    }

    private void addMonitEndStepExtTrace() {
        if (!isInOnlySqlOperation()) {
            FlowAttributes flowAttributes = PetalsExecutionContext.getFlowAttributes();
            String flowInstanceId = "";
            String flowStepId = "";
            if (flowAttributes != null) {
                flowInstanceId = flowAttributes.getFlowInstanceId();
                flowStepId = flowAttributes.getFlowStepId();
            }
            logger.log(Level.MONIT, "",
                    new ProvideExtFlowStepEndLogData(flowInstanceId, flowStepId));
        }
    }

    private boolean isInOnlySqlOperation() {
        QName sqlOperation = QName.valueOf((String) ExecutionContext.getProperties().get(
                PetalsExecutionContext.FLOW_STEP_OPERATION_NAME_PROPERTY_NAME));
        return Constants.OPERATION_INSERT.equals(sqlOperation.getLocalPart())
                || Constants.OPERATION_DELETE.equals(sqlOperation.getLocalPart())
                || Constants.OPERATION_UPDATE.equals(sqlOperation.getLocalPart());
    }
}
